from django.utils.functional import lazy
from django.utils.safestring import mark_safe


def is_object_creator(object, profile):
    return object.profile == profile


# deprecated
# https://docs.djangoproject.com/en/3.2/topics/i18n/translation/#other-uses-of-lazy-in-delayed-translations
# https://stackoverflow.com/questions/10706296/how-do-i-input-html-into-the-help-text-of-a-django-form-field
# obsolete as of django v4.1; https://docs.djangoproject.com/en/4.1/releases/4.1/#utilities
mark_safe_lazy = lazy(mark_safe, str)



from django.contrib.auth.models import User
from django.db.models.signals import post_save
from django.dispatch import receiver

from django_util.models import Profile


@receiver(post_save, sender=User)
def create_user_profile(sender, instance, created, **kwargs):
    """
    When User object is created, create Profile object
    """
    if created:
        Profile.objects.create(user=instance)
    else:
        instance.profile.save()



from django.contrib.auth.models import User
from django.test import TestCase

from django_util.models import Profile


# Create your tests here.
def user_example1(**kwargs):
    user = User.objects.create(username="your_username", password="your_password")
    user.save()
    return user


class UserModelTests(TestCase):
    """
    Test User model
    """

    def test_user(self):
        u1 = user_example1()
        self.assertTrue(isinstance(u1, User))
        self.assertEqual(u1.username, "your_username")
        self.assertEqual(u1.password, "your_password")


class ProfileModelTests(TestCase):
    """
    Test Profile model
    """

    def test_profile(self):
        u1 = user_example1()
        p1 = Profile.objects.get(id=1)
        self.assertTrue(isinstance(p1, Profile))
        self.assertEqual(u1.profile, p1)
        self.assertEqual(p1.balance, 0)



from typing import Any
from urllib.parse import urlparse

from django.contrib.contenttypes.fields import GenericRelation
from django.contrib.postgres.fields import ArrayField
from django.db import models
from django.db.models import ManyToManyField, ManyToManyRel, ManyToOneRel

# general
related_fields = (
    GenericRelation,
    ManyToManyField,
    ManyToManyRel,
    ManyToOneRel,
)

complex_fields = (ArrayField,)


# class
class UpperTextField(models.TextField):
    """
    A TextField that is stripped and uppercase
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    def get_prep_value(self, value: Any) -> Any:
        if value:
            return str(value).strip().upper()
        else:
            return value


class NoParamURLField(models.URLField):
    """
    A URLField that parse and keep only netloc and path
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    def get_prep_value(self, value: Any) -> Any:
        if value is not None:
            tmp = urlparse(value)
            output = tmp.netloc + tmp.path
            return output
        else:
            return None


# for admin.py
list_display_exclude = tuple(list(related_fields) + list(complex_fields))
list_filter_base = ["is_admin_verified"]
list_filter_base_public_contribute = list_filter_base + ["is_verified"]



import uuid

from django.contrib.auth.models import User
from django.contrib.postgres.fields import ArrayField
from django.core.validators import MinValueValidator
from django.db import models

from django_util.choices import (
    PersonBloodChoices,
    PersonEyeColorChoices,
    PersonGenderChoices,
    PersonRaceChoices,
)
from django_util.fields import UpperTextField, complex_fields, related_fields


# abstract
class Profile(models.Model):
    """
    Profile model extend Django User model

    Reference
    ---------
    - https://simpleisbetterthancomplex.com/tutorial/2016/07/22/how-to-extend-django-user-model.html#onetoone
    """

    user = models.OneToOneField(
        User,
        on_delete=models.CASCADE,
    )

    class Meta:
        abstract = True


class Base(models.Model):
    """
    The ideal abstract base model

    Reference
    ---------
    - https://docs.djangoproject.com/en/3.0/topics/db/models/#abstract-base-classes
    - https://stackoverflow.com/a/16838663

    Metadata

    - https://dataedo.com/kb/data-glossary/what-is-metadata

    User tracker

    - https://docs.djangoproject.com/en/3.1/topics/class-based-views/generic-editing/#models-and-request-user
    - https://stackoverflow.com/questions/5611410/related-name-argument-not-working-as-expected-in-django-model
    """

    created_at = models.DateTimeField(
        auto_now_add=True,
    )
    updated_at = models.DateTimeField(
        auto_now=True,
    )
    # The field ConcreteModel.created_by was declared with a lazy reference to 'django_util.profile', but app 'django_util' doesn't provide model 'profile'.
    # The field ConcreteModel.updated_by was declared with a lazy reference to 'django_util.profile', but app 'django_util' doesn't provide model 'profile'.
    # created_by = models.ForeignKey(
    # updated_by = models.ForeignKey(
    #     Profile,
    #     editable=False,
    #     null=True,
    #     on_delete=models.SET_NULL,
    #     related_name="%(class)s_created_by",
    # )
    note = UpperTextField(
        blank=True,
        default="",
    )
    is_admin_verified = models.BooleanField(
        default=False,
        editable=False,
    )

    class Meta:
        abstract = True

    @staticmethod
    def get_base_meta_field_list() -> list:
        return sorted(
            [
                "created_at",
                "updated_at",
            ]
        )

    @staticmethod
    def get_extra_meta_field_list() -> list:
        return sorted(
            [
                # "created_by",
                # "updated_by",
                "note",
                "is_admin_verified",
            ]
        )

    @classmethod
    def get_field_list(cls):
        return sorted([field.name for field in cls._meta.get_fields()])

    @classmethod
    def get_field_list_without_extra_meta(cls):
        return sorted(
            list(set(cls.get_field_list()) - set(cls.get_extra_meta_field_list()))
        )

    @classmethod
    def get_field_list_without_meta(cls):
        return sorted(
            list(
                set(cls.get_field_list())
                - set(cls.get_base_meta_field_list())
                - set(cls.get_extra_meta_field_list())
            )
        )

    @classmethod
    def get_related_field_list(cls):
        field_list = []
        for field in cls._meta.get_fields():
            if isinstance(field, related_fields):
                field_list.append(field.name)
        return sorted(field_list)

    @classmethod
    def get_complex_field_list(cls):
        field_list = []
        for field in cls._meta.get_fields():
            if isinstance(field, complex_fields):
                field_list.append(field.name)
        return sorted(field_list)


class BasePublicContribute(Base):
    """
    Enhanced abstract base model, use if data can be edited by public
    """

    is_verified = models.BooleanField(
        default=False,
    )
    is_hidden = models.BooleanField(
        default=False,
    )

    class Meta:
        abstract = True

    @staticmethod
    def get_staff_only_field_list():
        return sorted(
            [
                "is_hidden",
                "is_verified",
            ]
        )


class EmailHttpRequest(models.Model):
    """
    Abstract model that stores email and HttpRequest fields
    """

    email = models.EmailField(
        db_index=True,
    )
    campaign = UpperTextField(
        blank=True,
        default="",
    )
    http_referrer = UpperTextField(
        blank=True,
        default="",
    )
    http_user_agent = UpperTextField(
        blank=True,
        default="",
    )
    remote_addr = models.GenericIPAddressField(
        blank=True,
        null=True,
    )
    remote_host = UpperTextField(
        blank=True,
        default="",
    )

    class Meta:
        abstract = True


class Person(models.Model):
    """
    Abstract Person model

    Reference
    ---------
    - https://schema.org/Person
    """

    uuid = models.UUIDField(
        db_index=True,
        default=uuid.uuid4,
        editable=False,
        unique=True,
    )

    english_first_name = UpperTextField(
        default="",
    )
    english_middle_name = UpperTextField(
        blank=True,
        default="",
    )
    english_last_name = UpperTextField(
        default="",
    )
    native_first_name = UpperTextField(
        blank=True,
        default="",
    )
    native_middle_name = UpperTextField(
        blank=True,
        default="",
    )
    native_last_name = UpperTextField(
        blank=True,
        default="",
    )
    alias = ArrayField(
        UpperTextField(),
        blank=True,
        null=True,
    )
    description = UpperTextField(
        blank=True,
        default="",
    )

    # biology
    birth_date = models.DateTimeField(
        blank=True,
        null=True,
    )
    death_date = models.DateTimeField(
        blank=True,
        null=True,
    )
    blood_type = UpperTextField(
        blank=True,
        choices=PersonBloodChoices.choices,
        default=PersonBloodChoices.DEFAULT,
    )
    eye_color = UpperTextField(
        blank=True,
        choices=PersonEyeColorChoices.choices,
        default=PersonEyeColorChoices.DEFAULT,
    )
    gender = UpperTextField(
        blank=True,
        choices=PersonGenderChoices.choices,
        default=PersonGenderChoices.DEFAULT,
    )
    race = UpperTextField(
        blank=True,
        choices=PersonRaceChoices.choices,
        default=PersonRaceChoices.DEFAULT,
    )
    height = models.PositiveSmallIntegerField(
        blank=True,
        null=True,
        validators=[MinValueValidator(1)],
    )

    class Meta:
        abstract = True


class FullCalendarEventParseV6(models.Model):
    """
    FullCalendar Event

    Reference
    ---------
    - https://fullcalendar.io/docs/v6/event-parsing
    """

    group_id = models.TextField(
        blank=True,
        default="",
    )
    all_day = models.BooleanField(
        blank=True,
        null=True,
    )
    start = models.DateTimeField(
        blank=True,
        null=True,
    )
    end = models.DateTimeField(
        blank=True,
        null=True,
    )
    days_of_week = ArrayField(
        models.TextField(),
        blank=True,
        null=True,
    )
    start_time = models.TimeField(
        blank=True,
        null=True,
    )
    end_time = models.TimeField(
        blank=True,
        null=True,
    )
    start_recur = models.DateTimeField(
        blank=True,
        null=True,
    )
    end_recur = models.DateTimeField(
        blank=True,
        null=True,
    )
    title = models.TextField(
        blank=True,
        default="",
    )
    url = models.URLField(
        blank=True,
        default="",
    )
    interactive = models.BooleanField(
        blank=True,
        null=True,
    )
    class_name = ArrayField(
        models.TextField(),
        blank=True,
        null=True,
    )
    editable = models.BooleanField(
        blank=True,
        null=True,
    )
    start_editable = models.BooleanField(
        blank=True,
        null=True,
    )
    duration_editable = models.BooleanField(
        blank=True,
        null=True,
    )
    resource_editable = models.BooleanField(
        blank=True,
        null=True,
    )
    resource_id = models.TextField(
        blank=True,
        default="",
    )
    resource_ids = ArrayField(
        models.TextField(),
        blank=True,
        null=True,
    )
    display = models.TextField(
        blank=True,
        default="",
    )
    overlap = models.BooleanField(
        blank=True,
        null=True,
    )
    constraint = models.BooleanField(
        blank=True,
        null=True,
    )
    color = models.BooleanField(
        blank=True,
        null=True,
    )

    class Meta:
        abstract = True


class WiseWebhook(models.Model):
    """
    Wise Webhook

    Reference
    ---------
    - https://docs.wise.com/api-docs/api-reference/webhook
    """

    data = models.JSONField(
        blank=True,
        null=True,
    )
    subscription_id = models.TextField(
        blank=True,
        default="",
    )
    event_type = models.TextField(
        blank=True,
        default="",
    )
    schema_version = models.TextField(
        blank=True,
        default="",
    )
    sent_at = models.DateTimeField(
        blank=True,
        null=True,
    )

    class Meta:
        abstract = True



from django import template
from django.templatetags.static import static
from django.urls import reverse

register = template.Library()


# filter
@register.filter
def get_item(dictionary, key):
    """
    Get dictionary value based on key
    """
    return dictionary.get(key)


@register.filter
def index(indexable, i):
    """
    Get index value of indexable
    """
    return indexable[i]


@register.filter
def sentence_case(str_):
    """
    Convert to sentence case
    """
    if str_:
        return ". ".join([l.capitalize() for l in str_.split(". ")])
    else:
        return


@register.filter
def list_to_newline(list_):
    """
    Convert list into string with newline
    """
    if isinstance(list_, list) and list_:
        return "\n".join(list_)
    else:
        return "---"


@register.filter
def ifinlist(value, list_):
    """
    Check if value is in list
    """
    return value in list_


@register.filter
def as_abs_url(path, request):
    """
    Build Absolute url

    Reference
    ---------
    - https://stackoverflow.com/questions/19024159/how-to-reverse-a-name-to-a-absolute-url-in-django-template/34630561
    """
    return request.build_absolute_uri(path)


@register.filter
def get_image_url(queryset, i):
    """
    Get image queryset specific index

    Reference
    ---------
    - https://stackoverflow.com/questions/4651172/reference-list-item-by-index-within-django-template
    - https://stackoverflow.com/questions/28286207/reference-static-files-in-django-in-py-files-not-in-templates
    """
    try:
        return queryset[int(i)].image.url
    except IndexError:
        return static("img/no_image_available.svg")
    else:
        return


@register.filter
def social_media_name_mapper(str_):
    hashmap = {
        "crunchbase": "crunchbase",
        "facebook": "facebook",
        "instagram": "instagram",
        "linkedin": "linkedin",
        "tiktok": "tiktok",
        "twitter": "twitter",
        "wikipedia": "wikipedia",
        "youtube": "youtube",
    }
    return hashmap.get(str_, None)


# tags
@register.simple_tag
def combine_url_lists(list1, list2):
    """
    Combine two provided lists
    """
    return list1 + list2


@register.simple_tag(takes_context=True)
def abs_url(context, view_name, *args, **kwargs):
    """
    Build absolute url

    Reference
    ---------
    - https://stackoverflow.com/questions/19024159/how-to-reverse-a-name-to-a-absolute-url-in-django-template/34630561
    """
    # Could add except for KeyError, if rendering the template
    # without a request available.
    return context["request"].build_absolute_uri(
        reverse(
            view_name,
            args=args,
            kwargs=kwargs,
        )
    )


@register.simple_tag(takes_context=True)
def query_transform(context, **kwargs):
    """
    Create page pagination

    Reference
    ---------
    - https://docs.djangoproject.com/en/3.2/howto/custom-template-tags/
    - https://stackoverflow.com/questions/3324009/how-to-obtain-all-values-of-a-multi-valued-key-from-djangos-request-get-querydi
    - https://stackoverflow.com/questions/46026268/pagination-and-get-parameters
    """
    query = context["request"].GET.copy()
    for k, v in kwargs.items():
        query[k] = v
    # print(query)
    # print(query.urlencode())
    return query.urlencode()



from setuptools import find_packages, setup

import versioneer

with open("README.md", "r") as f:
    LONG_DESCRIPTION = f.read()
with open("requirements.txt", "r") as f:
    REQUIREMENTS = [line.strip() for line in f]

CLASSIFIERS = [
    "Environment :: Web Environment",
    "Framework :: Django",
    "Framework :: Django :: 4.2",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: BSD License",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3 :: Only",
    "Programming Language :: Python :: 3.11",
    "Topic :: Internet :: WWW/HTTP",
    "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
]

setup(
    author_email="shawn.coding.acc@gmail.com",
    author="Shawn Ng",
    classifiers=CLASSIFIERS,
    cmdclass=versioneer.get_cmdclass(),
    description="A Django application that provides generic utility.",
    install_requires=REQUIREMENTS,
    license="BSD-3",
    long_description=LONG_DESCRIPTION,
    name="django-util",
    packages=find_packages(include=["django_util", "django_util.*"]),
    platforms="any",
    test_suite="nose.collector",
    tests_require=["nose"],
    url="https://github.com/shawnngtq/django-util",
    version=versioneer.get_version(),
)



from django.db import models
from django.utils.translation import gettext_lazy as _

TRUE_FALSE_CHOICES = [
    ("", "---"),
    (True, "TRUE"),
    (False, "FALSE"),
]


class PersonBloodChoices(models.TextChoices):
    """
    Person blood enumeration
    """

    DEFAULT = "", _("---")
    A = "A", _("A")
    B = "B", _("B")
    AB = "AB", _("AB")
    O = "O", _("O")
    OTHERS = "OTHERS", _("OTHERS")


class PersonEducationLevelChoices(models.TextChoices):
    """
    Person education level enumeration

    Reference
    ---------
    - https://study.com/different_degrees.html
    """

    DEFAULT = "", _("---")
    HIGH_SCHOOL = "HIGH_SCHOOL", _("HIGH SCHOOL")
    ASSOCIATE = "ASSOCIATE", _("ASSOCIATE DEGREE")
    BACHELOR = "BACHELOR", _("BACHELOR'S DEGREE")
    MASTER = "MASTER", _("MASTER'S DEGREE")
    DOCTORAL = "DOCTORAL", _("DOCTORAL DEGREE")
    OTHERS = "OTHERS", _("OTHERS")


class PersonEyeColorChoices(models.TextChoices):
    """
    Person eye color enumeration

    Reference
    ---------
    - https://en.wikipedia.org/wiki/Eye_color#Eye_color_chart_(Martin_scale)
    """

    DEFAULT = "", _("---")
    AMBER = "AMBER", _("AMBER")
    BLUE = "BLUE", _("BLUE")
    BROWN = "BROWN", _("BROWN")
    GREEN = "GREEN", _("GREEN")
    GREY = "GREY", _("GREY")
    HAZEL = "HAZEL", _("HAZEL")
    RED = "RED", _("RED")
    OTHERS = "OTHERS", _("OTHERS")


class PersonGenderChoices(models.TextChoices):
    """
    Person gender enumeration
    """

    DEFAULT = "", _("---")
    FEMALE = "FEMALE", _("FEMALE")
    MALE = "MALE", _("MALE")
    OTHERS = "OTHERS", _("OTHERS")


class PersonRaceChoices(models.TextChoices):
    """
    Person race enumeration

    Reference
    ---------
    - https://grants.nih.gov/grants/guide/notice-files/not-od-15-089.html
    """

    DEFAULT = "", _("---")
    ASIAN = "ASIAN", _("ASIAN")
    BLACK = "BLACK", _("BLACK OR AFRICAN AMERICAN")
    HISPANIC = "HISPANIC", _("HISPANIC OR LATINO")
    INDIAN = "INDIAN", _("AMERICAN INDIAN OR ALASKA NATIVE")
    ISLANDER = "ISLANDER", _("NATIVE HAWAIIAN OR OTHER PACIFIC ISLANDER")
    WHITE = "WHITE", _("WHITE")
    OTHERS = "OTHERS", _("OTHERS")
